/*
	hex pattern finder v2.2
	additional features:
	- file specification:
		- wildcards in file spec
		- recursive search in subdirectories
	- search pattern:
		- wildcards allowed
		- text strings, optional case insensitive

	(c) 1998 - 2000 Heinz Repp
		2003-07-03 Linux port
*/

#define _GNU_SOURCE

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fnmatch.h>
#include <dirent.h>

#define FALSE 0
#define TRUE 1
#define MAX_PATH 1024
#define MAX_FILENAME 256
#define BUFFERSIZE 65536  /* 64 kB */

short int subdirs = FALSE, verbose = FALSE, no_case = FALSE, symlinks = FALSE,
	hidden = FALSE, usage = FALSE, p_length, p_signif, jumpm,
	shift[256], jump[256], p_posit[256], letter[256],
	ch_prob [256] = {
14904, 1533, 1038, 865, 936, 615, 614, 643, 738, 450, 761, 389, 493, 602, 391, 528,
 602, 339, 282, 251, 346, 262, 285, 217, 300, 212, 211, 218, 277, 209, 262, 227,
 2663, 227, 320, 222, 532, 226, 342, 199, 314, 263, 301, 231, 344, 310, 366, 295,
 638, 409, 357, 414, 323, 297, 297, 252, 311, 293, 257, 248, 233, 272, 215, 416,
 567, 516, 313, 442, 587, 725, 513, 300, 296, 368, 171, 187, 373, 399, 351, 387,
 673, 235, 387, 442, 414, 362, 349, 301, 233, 172, 184, 184, 256, 208, 256, 444,
 247, 778, 317, 472, 558, 1365, 630, 349, 486, 783, 319, 210, 618, 397, 804, 722,
 501, 158, 810, 692, 1156, 583, 354, 692, 286, 253, 202, 162, 230, 239, 307, 275,
 722, 326, 244, 578, 254, 306, 200, 186, 685, 486, 201, 1045, 191, 366, 205, 171,
 253, 164, 167, 166, 154, 130, 140, 127, 163, 197, 234, 144, 144, 127, 130, 142,
 203, 188, 131, 140, 144, 136, 124, 120, 146, 117, 148, 117, 139, 119, 127, 127,
 193, 133, 129, 118, 146, 124, 135, 131, 228, 141, 140, 211, 142, 126, 139, 198,
 509, 292, 217, 229, 356, 126, 181, 283, 183, 178, 142, 136, 352, 133, 139, 138,
 202, 152, 158, 135, 140, 118, 131, 153, 188, 138, 133, 144, 159, 155, 127, 141,
 234, 149, 142, 136, 171, 137, 130, 127, 477, 263, 134, 221, 256, 127, 180, 175,
 371, 188, 181, 163, 206, 154, 212, 273, 355, 184, 230, 218, 384, 216, 364, 3928
 }; /* character probability in 1/100000 */
unsigned long filecount = 0L, bytecount = 0L, matchcount = 0L;
unsigned char buffer[BUFFERSIZE + 511], poschar[256];

char nibble (char c)
{
	c = tolower (c);
	if (c < '0' || (c > '9' && c < 'a') || c > 'f')
	{
		fputs ("error: invalid hexadecimal pattern\n", stderr);
		exit (2);
	}

	return c < 'a' ? c - '0' : c - ('a' - 10);
} /* nibble */


void ParseOption (char *option)
{
	do
	{
		switch (tolower (*option))
		{
		case 'h':
			switch (*++option)
			{
			case '-':
				option++;
				hidden = FALSE;
				break;
			case '+':
				option++;
			default:
				hidden = TRUE;
			} /* endswitch h */
			break;

		case 'i':
			switch (*++option)
			{
			case '-':
				option++;
				no_case = FALSE;
				break;
			case '+':
				option++;
			default:
				no_case = TRUE;
			} /* endswitch i */
			break;

		case 'r':
			switch (*++option)
			{
			case '-':
				option++;
				subdirs = FALSE;
				break;
			case '+':
				option++;
			default:
				subdirs = TRUE;
			} /* endswitch r */
			break;

		case 's':
			switch (*++option)
			{
			case '-':
				option++;
				symlinks = FALSE;
				break;
			case '+':
				option++;
			default:
				symlinks = TRUE;
			} /* endswitch l */
			break;

		case 'v':
			switch (*++option)
			{
			case '-':
				option++;
				verbose = FALSE;
				break;
			case '+':
				option++;
			default:
				verbose = TRUE;
			} /* endswitch r */
			break;

		case '?':
			usage = TRUE;
			return;

		default:
			fprintf (stderr, "error: unknown option '%s'\n", option);
			exit (3);

		} /* endswitch *option */
	} while (*option);

} /* ParseOption */


void ParsePattern (char *hexarg)
{
	register int i, j;
	int          next_is, lastw;
	short int    pattern[256], p_lttr[256];

	p_length = 0;
	p_signif = 0;
	lastw = 0;
	next_is = 0;

	do
	{
		switch (next_is)
		{
		case 0:		// new pattern
			switch (*hexarg)
			{
			case '/':		/* string delimiter */
				hexarg++;
				next_is = 2;
				break;
			case ' ':
				hexarg++;	/* ignore blanks */
				break;
			case '?':		 /* wildcard */
				hexarg++;
				pattern[p_length] = -1;
				lastw = ++p_length;
				break;
			default:
				pattern[p_length] = nibble (*hexarg++) << 4;
				next_is = 1;
			}
			break;
		case 1:		// second nibble expected
			pattern[p_length] |= nibble (*hexarg++);
			if (no_case)
				p_lttr[p_length] = FALSE;
			p_posit[p_signif++] = p_length++;
			next_is = 0;
			break;
		case 2:		// string processed
			if (*hexarg == '/')
			{
				hexarg++;
				next_is = 0;
				break;
			}
			if (*hexarg == '\\')
			{
				hexarg++;
				next_is = 3;
				break;
			}
			/* if none of both (back)slashes: fall-through! */
		case 3:		// escaped character in string
			pattern[p_length] = *hexarg++;
			if (no_case)
			{
				if (pattern[p_length] >= 'A' && pattern[p_length] <= 'Z')
				{
					pattern[p_length] |= 0x20;
					p_lttr[p_length] = TRUE;
				}
				else if (pattern[p_length] >= 'a' && pattern[p_length] <= 'z')
					p_lttr[p_length] = TRUE;
				else
					p_lttr[p_length] = FALSE;
			}
			p_posit[p_signif++] = p_length++;
			next_is = 2;
			break;
		} /* endswitch next_is */

	} while (*hexarg && p_length < 256); /* enddo */

	if (next_is)
	{
		if (next_is > 1)
		{
			fputs ("error: missing string end ('/')\n", stderr);
			exit (4);
		}
		else
			nibble ('?');    /* causes appropriate error exit */
	}

	if (! p_signif)
	{
		fputs ("error: empty search pattern\n", stderr);
		exit (5);
	}

	if (lastw == p_length)
	{
		fputs ("warning: trailing wildcard(s) ignored\n", stderr);
		p_length--;
		while (pattern[--p_length] < 0);
		lastw = p_length++;
		while (lastw && pattern[lastw-1] >= 0)
			lastw--;
	}

	/* insertion sort p_posit up to p_signif-2 in ch_prob order */
	if (no_case)
		for (i = 'a'; i <= 'z'; i++)
			ch_prob[i] += ch_prob[i & ~0x20];

	for (i = 1; i < p_signif - 1; i++)
	{
		next_is = p_posit[i];
		for (j = i;
				 j > 0 && ch_prob[pattern[p_posit[j - 1]]] < ch_prob[pattern[next_is]];
				 j--) p_posit[j] = p_posit[j - 1];
		p_posit[j] = next_is;
	}

	/* make shift and jump tables */
	for (i = 0; i < 256; i++)
		 shift[i] = p_length - lastw;
	for (i = 0; i < p_signif; i++)
	{
		poschar[i] = pattern[p_posit[i]];
		if (no_case) letter[i] = p_lttr[p_posit[i]];

		if (p_posit[i] >= lastw)
		{
			shift[poschar[i]] = p_length - 1 - p_posit[i];
			if (no_case && letter[i])
				shift[poschar[i] & ~0x20] = shift[poschar[i]];
		}
		jump[i] = p_length;
	}
	jumpm = p_length;
	/* try decreasing jumps after (partial) match */
	for (j = p_length - 1; j > 0; j--)
	{
		/* check decreasing char positions:
			if match is possible on shifted pattern after:
			- mismatch, note shift distance as possible
			- match before, try next position           */
		for (i = p_signif - 1; i >= 0; i--)
		{
			if (p_posit[i] < j || pattern[p_posit[i] - j] < 0)
				jump[i] = j;
			else if (no_case && letter[i] != p_lttr[p_posit[i] - j])
			{
				if (letter[i])
				{
					if (poschar[i] != (pattern[p_posit[i] - j] | 0x20))
					{
						jump[i] = j;
						break;
					}
				}
				else
				{
					jump[i] = j;
					if ((poschar[i] | 0x20) != pattern[p_posit[i] - j])
						break;
				}
			}
			else if (poschar[i] != pattern[p_posit[i] - j])
			{
				jump[i] = j;
				break;
			}
		}
		if (i < 0)
			jumpm = j;
	}

	if (verbose)
	{
		printf ("searching for hex pattern ");
		for (i = 0; i < p_length; i++)
			if (pattern[i] < 0)
				printf ("?? ");
			else
				printf ("%02x ", pattern[i]);
		printf ("...\n");
	}

} /* ParsePattern */


unsigned char *FastSearch (register unsigned char *where)
{
	register int i;

	/* outer loop */
	for ( ; ; )
	{
		where += p_length - 1;

		/* inner loop */
		while ((i = shift[*where]))
			where += i;
		/* inner loop */

		where -= p_length - 1;

		i = p_signif - 1;
		do
			if (--i < 0)
				return where;
		while (no_case && letter[i] ?
			(where[p_posit[i]] | 0x20) == poschar[i] :
			where[p_posit[i]] == poschar[i]);

		where += jump[i];
	} /* outer loop */

} /* FastSearch */


int SearchFile (char *filename)
{
	unsigned long result, f_offset;
	unsigned char *fill, *found;
	char pdc[12];
	int handle, i;

	if (filename)
	{
		if ((handle = open (filename, O_RDONLY, 0)) < 0)
		{
			fprintf (stderr, "error: could not open '%s'\n", filename);
			return FALSE;
		}
	}
	else
	{
		handle = 0;
		filename = "stdin";
	}

	if (verbose)
		printf ("searching %s ...\n", filename);

	f_offset = 0;
	fill = buffer;
	pdc[0] = '(';

	while ((result = read (handle, fill, BUFFERSIZE)) > 0)
	{
		while (result < BUFFERSIZE)
			if ((i = read (handle, fill+result, BUFFERSIZE-result)))
				result += i;
			else
				break;

		/* statistics */
		bytecount += result;

		/* append pattern after end */
		fill += result;
		for (i = 0; i < p_signif; i++)
			fill[p_posit[i]] = poschar[i];

		found = FastSearch (buffer);

		while (found <= fill - p_length)
		{
			/* it's a real match - print and count it */
			sprintf (&pdc[1], "%lu", f_offset + (found - buffer));
			printf ("offset 0x%08lX%11s) in %s\n",
							f_offset + (found - buffer), pdc, filename);
			matchcount++;

			found = FastSearch (found + jumpm);
		}

		if (result < BUFFERSIZE)
			break;
		else
		{
			result = p_length - 1;
			memcpy (buffer, fill - result, result);
			f_offset += fill - buffer - result;
			fill = buffer + result;
		}
	}

	filecount++;
	close (handle);
	return TRUE;
} /* SearchFile */


int SearchPath (char *path, char *mask)
{
	int found;
	DIR *CurrentDir;
	struct dirent *CurrentEntry;
	char *FullName;
	struct stat stbuff;

	FullName = malloc (strlen(path)+MAX_FILENAME+2);
	found = FALSE;
	if ((CurrentDir = opendir (path)))
	{
		while ((CurrentEntry = readdir (CurrentDir)))
		{
			if ((symlinks || CurrentEntry->d_type != DT_LNK) &&
				(hidden || CurrentEntry->d_name[0] != '.'))
			{
				strcpy (FullName, path);
				strcat (FullName, "/");
				strncat (FullName, CurrentEntry->d_name, MAX_FILENAME);
				if (CurrentEntry->d_type == DT_LNK &&
					stat (FullName, &stbuff) == 0)
					if (S_ISREG (stbuff.st_mode))
						CurrentEntry->d_type = DT_REG;
					else if (S_ISDIR (stbuff.st_mode))
						CurrentEntry->d_type = DT_DIR;

				if (CurrentEntry->d_type == DT_REG)	// regular file
				{
					if (! fnmatch (mask, CurrentEntry->d_name, no_case ? FNM_CASEFOLD : 0))
							found |= SearchFile (FullName);
				}
				else if (subdirs && CurrentEntry->d_type == DT_DIR)	// directory
				{
					if (strcmp(CurrentEntry->d_name, ".") && strcmp(CurrentEntry->d_name, ".."))
						found |= SearchPath (FullName, mask);
				}
			}
		}
		closedir (CurrentDir);
	}
	free (FullName);
	return found;
} /* SearchPath */


int main (int argc, char *argv[])
{
	int i, found = TRUE;
	char *filespec, *wildcard;
	struct stat stbuff;

	/* parse command line: */

	/* 1. leading options */
	i = 1;
	while (i < argc && *argv[i] == '-')
		ParseOption (++argv[i++]);

	if (verbose)
		printf ("hexfind 2.2 for Linux (c) 1998-2003 by Heinz Repp\n");

	if (usage)
	{
		printf (
			"usage:   %s [-irhsv[+|-]?] pattern [files ...]\n"
			"   options:  <i>gnore case, <r>ecurse into subdirectories,\n"
			"             include <h>idden (dot) files & dirs,\n"
			"             follow <s>ymlinks, be <v>erbose\n"
			"             options may appear anywhere, order is significant,\n"
			"             a trailing '-' disables, '+' enables (default)\n"
			"   pattern:  one or more of [ HEX-pair | /string/ | ? ]\n"
			"             (? = wildcard, pattern must be contiguous or quoted!)\n",
			argv[0]);
		exit (0);
	}

	/* 2. pattern */
	if (i >= argc)
	{
		fprintf (stderr, "error: missing search pattern\n"
			"type '%s -?' for help\n", argv[0]);
		exit (1);
	}
	ParsePattern (argv[i]);

	/* 3. trailing options and filespecs */
	while (++i < argc)
	{
		if (*argv[i] == '-')
			ParseOption (++argv[i]);
		else
		{
			found = FALSE;
			/* amy wildcards? */
			if ((wildcard = strpbrk (argv[i], "?*")))
			{
				if ((filespec = strrchr (argv[i], '/')))
				{
					if (filespec > wildcard)
					{
						fprintf (stderr, "bad argument (wildcard in path): discarding %s\n", argv[i]);
						continue;
					}
					else
					{
						*filespec++ = '\0';
						found = SearchPath (argv[i], filespec);
						*--filespec = '/';
					}
				}
				else
				{
					found = SearchPath (".", argv[i]);
				}
			}
			else
			{
				/* no special treatment for hidden files, as pathname expansion of
				the shell usually ignores leading '.' and '.' and '..' files */
				if (symlinks)
					found = ! stat (argv[i], &stbuff);
				else
					found = ! lstat (argv[i], &stbuff);

				if (found)
					if (S_ISDIR (stbuff.st_mode))
						found = SearchPath (argv[i], "*");
					else if (S_ISREG (stbuff.st_mode))
						found = SearchFile (argv[i]);
					else
						continue;	/* ignore silently all others */
			}
			if (! found)
				fprintf (stderr, "found no file(s) matching '%s'\n", argv[i]);

		} /* endif */
	} /* endwhile all arguments */

	/* if no file spec until now */
	if (found && ! filecount)
		SearchFile (NULL);

	if (verbose)
		printf ("found %lu occurence(s) in %lu file(s) (%lu bytes searched)\n",
			matchcount, filecount, bytecount);

	exit (0);
} /* main */
